#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import unittest

from openfermion.hamiltonians import MolecularData
from openfermion.ops import FermionOperator, QubitOperator
from openfermion.transforms import (binary_code_transform, bravyi_kitaev,
                                    get_fermion_operator, jordan_wigner)
from openfermion.utils import eigenspectrum

from openfermion.transforms._binary_codes import *


def lih_hamiltonian():
    geometry = [('Li', (0., 0., 0.)), ('H', (0., 0., 1.45))]
    active_space_start = 1
    active_space_stop = 3
    molecule = MolecularData(geometry, 'sto-3g', 1,
                             description="1.45")
    molecule.load()
    molecular_hamiltonian = molecule.get_molecular_hamiltonian(
        occupied_indices=range(active_space_start),
        active_indices=range(active_space_start, active_space_stop))
    hamiltonian = get_fermion_operator(molecular_hamiltonian)
    ground_state_energy = eigenspectrum(hamiltonian)[0]
    return hamiltonian, ground_state_energy


class LinearizeDecoderTest(unittest.TestCase):

    def test_linearize(self):
        a = linearize_decoder([[0, 1, 1], [1, 0, 0]])
        self.assertListEqual([str(a[0]), str(a[1])], ['[W1] + [W2]', '[W0]'])


class CodeTransformTest(unittest.TestCase):
    def test_tranform_function(self):
        ferm_op = FermionOperator('2')
        n_modes = 5
        qubit_op = binary_code_transform(ferm_op, parity_code(n_modes))
        correct_op = QubitOperator(((1, 'Z'), (2, 'X'), (3, 'X'), (4, 'X')),
                                   0.5) + \
                     QubitOperator(((2, 'Y'), (3, 'X'), (4, 'X')), 0.5j)
        self.assertTrue(qubit_op == correct_op)
        ferm_op = FermionOperator('2^')
        n_modes = 5
        qubit_op = binary_code_transform(ferm_op, parity_code(n_modes))
        correct_op = QubitOperator(((1, 'Z'), (2, 'X'), (3, 'X'), (4, 'X')),
                                   0.5) \
                     + QubitOperator(((2, 'Y'), (3, 'X'), (4, 'X')), -0.5j)
        self.assertTrue(qubit_op == correct_op)

        ferm_op = FermionOperator('5^')
        op2 = QubitOperator('Z0 Z1 Z2 Z3 Z4 X5', 0.5) \
              - QubitOperator('Z0 Z1 Z2 Z3 Z4 Y5', 0.5j)
        op1 = binary_code_transform(ferm_op, jordan_wigner_code(6))
        self.assertTrue(op1 == op2)

    def test_checksum_code(self):
        hamiltonian, gs_energy = lih_hamiltonian()
        code = checksum_code(4, 0)
        qubit_hamiltonian = binary_code_transform(hamiltonian, code)
        self.assertAlmostEqual(gs_energy,
                               eigenspectrum(qubit_hamiltonian)[0])

    def test_jordan_wigner(self):
        hamiltonian, gs_energy = lih_hamiltonian()
        code = jordan_wigner_code(4)
        qubit_hamiltonian = binary_code_transform(hamiltonian, code)
        self.assertAlmostEqual(gs_energy,
                               eigenspectrum(qubit_hamiltonian)[0])
        self.assertDictEqual(qubit_hamiltonian.terms,
                             jordan_wigner(hamiltonian).terms)

    def test_bravyi_kitaev(self):
        hamiltonian, gs_energy = lih_hamiltonian()
        code = bravyi_kitaev_code(4)
        qubit_hamiltonian = binary_code_transform(hamiltonian, code)
        self.assertAlmostEqual(gs_energy,
                               eigenspectrum(qubit_hamiltonian)[0])
        qubit_spectrum = eigenspectrum(qubit_hamiltonian)
        fenwick_spectrum = eigenspectrum(bravyi_kitaev(hamiltonian))
        for eigen_idx, eigenvalue in enumerate(qubit_spectrum):
            self.assertAlmostEqual(eigenvalue, fenwick_spectrum[eigen_idx])

    def test_parity_code(self):
        hamiltonian, gs_energy = lih_hamiltonian()
        code = parity_code(4)
        qubit_hamiltonian = binary_code_transform(hamiltonian, code)
        self.assertAlmostEqual(gs_energy,
                               eigenspectrum(qubit_hamiltonian)[0])

    def test_weight_one_binary_addressing_code(self):
        hamiltonian, gs_energy = lih_hamiltonian()
        code = interleaved_code(8) * (
                2 * weight_one_binary_addressing_code(2))
        qubit_hamiltonian = binary_code_transform(hamiltonian, code)
        self.assertAlmostEqual(gs_energy,
                               eigenspectrum(qubit_hamiltonian)[0])

    def test_weight_one_segment_code(self):
        hamiltonian, gs_energy = lih_hamiltonian()
        code = interleaved_code(6) * (2 * weight_one_segment_code())
        qubit_hamiltonian = binary_code_transform(hamiltonian, code)
        self.assertAlmostEqual(gs_energy,
                               eigenspectrum(qubit_hamiltonian)[0])

    def test_weight_two_segment_code(self):
        hamiltonian, gs_energy = lih_hamiltonian()
        code = weight_two_segment_code()
        qubit_hamiltonian = binary_code_transform(hamiltonian, code)
        self.assertAlmostEqual(gs_energy,
                               eigenspectrum(qubit_hamiltonian)[0])

    def test_interleaved_code(self):
        hamiltonian, gs_energy = lih_hamiltonian()
        code = interleaved_code(4)
        qubit_hamiltonian = binary_code_transform(hamiltonian, code)
        self.assertAlmostEqual(gs_energy,
                               eigenspectrum(qubit_hamiltonian)[0])



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Tests for quadratic_hamiltonian.py."""
from __future__ import absolute_import

import numpy
import unittest

from openfermion.config import EQ_TOLERANCE
from openfermion.ops import (normal_ordered,
                             FermionOperator, QuadraticHamiltonian)
from openfermion.ops._quadratic_hamiltonian import antisymmetric_canonical_form
from openfermion.transforms import get_fermion_operator, get_sparse_operator
from openfermion.utils import get_ground_state, majorana_operator
from openfermion.utils._testing_utils import (random_antisymmetric_matrix,
                                              random_hermitian_matrix)


class QuadraticHamiltoniansTest(unittest.TestCase):
    def setUp(self):
        self.n_qubits = 5
        self.constant = 1.7
        self.chemical_potential = 2.

        # Obtain random Hermitian and antisymmetric matrices
        self.hermitian_mat = random_hermitian_matrix(self.n_qubits)
        self.antisymmetric_mat = random_antisymmetric_matrix(self.n_qubits)

        self.combined_hermitian = (
            self.hermitian_mat -
            self.chemical_potential * numpy.eye(self.n_qubits))

        # Initialize a particle-number-conserving Hamiltonian
        self.quad_ham_pc = QuadraticHamiltonian(
            self.constant, self.hermitian_mat)

        # Initialize a non-particle-number-conserving Hamiltonian
        self.quad_ham_npc = QuadraticHamiltonian(
            self.constant, self.hermitian_mat, self.antisymmetric_mat,
            self.chemical_potential)

        # Initialize the sparse operators and get their ground energies
        self.quad_ham_pc_sparse = get_sparse_operator(self.quad_ham_pc)
        self.quad_ham_npc_sparse = get_sparse_operator(self.quad_ham_npc)

        self.pc_ground_energy, self.pc_ground_state = get_ground_state(
            self.quad_ham_pc_sparse)
        self.npc_ground_energy, self.npc_ground_state = get_ground_state(
            self.quad_ham_npc_sparse)

    def test_combined_hermitian_part(self):
        """Test getting the combined Hermitian part."""
        combined_hermitian_part = self.quad_ham_pc.combined_hermitian_part
        for i in numpy.ndindex(combined_hermitian_part.shape):
            self.assertAlmostEqual(self.hermitian_mat[i],
                                   combined_hermitian_part[i])

        combined_hermitian_part = self.quad_ham_npc.combined_hermitian_part
        for i in numpy.ndindex(combined_hermitian_part.shape):
            self.assertAlmostEqual(self.combined_hermitian[i],
                                   combined_hermitian_part[i])

    def test_hermitian_part(self):
        """Test getting the Hermitian part."""
        hermitian_part = self.quad_ham_pc.hermitian_part
        for i in numpy.ndindex(hermitian_part.shape):
            self.assertAlmostEqual(self.hermitian_mat[i], hermitian_part[i])

        hermitian_part = self.quad_ham_npc.hermitian_part
        for i in numpy.ndindex(hermitian_part.shape):
            self.assertAlmostEqual(self.hermitian_mat[i], hermitian_part[i])

    def test_antisymmetric_part(self):
        """Test getting the antisymmetric part."""
        antisymmetric_part = self.quad_ham_pc.antisymmetric_part
        for i in numpy.ndindex(antisymmetric_part.shape):
            self.assertAlmostEqual(0., antisymmetric_part[i])

        antisymmetric_part = self.quad_ham_npc.antisymmetric_part
        for i in numpy.ndindex(antisymmetric_part.shape):
            self.assertAlmostEqual(self.antisymmetric_mat[i],
                                   antisymmetric_part[i])

    def test_conserves_particle_number(self):
        """Test checking whether Hamiltonian conserves particle number."""
        self.assertTrue(self.quad_ham_pc.conserves_particle_number)
        self.assertFalse(self.quad_ham_npc.conserves_particle_number)

    def test_add_chemical_potential(self):
        """Test adding a chemical potential."""
        self.quad_ham_npc.add_chemical_potential(2.4)

        combined_hermitian_part = self.quad_ham_npc.combined_hermitian_part
        hermitian_part = self.quad_ham_npc.hermitian_part

        want_combined = (self.combined_hermitian -
                         2.4 * numpy.eye(self.n_qubits))
        want_hermitian = self.hermitian_mat

        for i in numpy.ndindex(combined_hermitian_part.shape):
            self.assertAlmostEqual(combined_hermitian_part[i],
                                   want_combined[i])

        for i in numpy.ndindex(hermitian_part.shape):
            self.assertAlmostEqual(hermitian_part[i], want_hermitian[i])

        self.assertAlmostEqual(2.4 + self.chemical_potential,
                               self.quad_ham_npc.chemical_potential)

    def test_orbital_energies(self):
        """Test getting the orbital energies."""
        # Test the particle-number-conserving case
        orbital_energies, constant = self.quad_ham_pc.orbital_energies()
        # Test the ground energy
        energy = numpy.sum(
            orbital_energies[orbital_energies < -EQ_TOLERANCE]) + constant
        self.assertAlmostEqual(energy, self.pc_ground_energy)

        # Test the non-particle-number-conserving case
        orbital_energies, constant = self.quad_ham_npc.orbital_energies()
        # Test the ground energy
        energy = constant
        self.assertAlmostEqual(energy, self.npc_ground_energy)

    def test_ground_energy(self):
        """Test getting the ground energy."""
        # Test particle-number-conserving case
        energy = self.quad_ham_pc.ground_energy()
        self.assertAlmostEqual(energy, self.pc_ground_energy)
        # Test non-particle-number-conserving case
        energy = self.quad_ham_npc.ground_energy()
        self.assertAlmostEqual(energy, self.npc_ground_energy)

    def test_majorana_form(self):
        """Test getting the Majorana form."""
        majorana_matrix, majorana_constant = self.quad_ham_npc.majorana_form()
        # Convert the Majorana form to a FermionOperator
        majorana_op = FermionOperator((), majorana_constant)
        normalization = 1. / numpy.sqrt(2.)
        for i in range(2 * self.n_qubits):
            if i < self.n_qubits:
                left_op = majorana_operator((i, 0), normalization)
            else:
                left_op = majorana_operator((i - self.n_qubits, 1),
                                            normalization)
            for j in range(2 * self.n_qubits):
                if j < self.n_qubits:
                    right_op = majorana_operator((j, 0),
                            majorana_matrix[i, j] * normalization)
                else:
                    right_op = majorana_operator((j - self.n_qubits, 1),
                            majorana_matrix[i, j] * normalization)
                majorana_op += .5j * left_op * right_op
        # Get FermionOperator for original Hamiltonian
        fermion_operator = normal_ordered(
            get_fermion_operator(self.quad_ham_npc))
        self.assertTrue(
            normal_ordered(majorana_op) == fermion_operator)

    def test_diagonalizing_bogoliubov_transform(self):
        """Test getting the diagonalizing Bogoliubov transformation."""
        hermitian_part = self.quad_ham_npc.combined_hermitian_part
        antisymmetric_part = self.quad_ham_npc.antisymmetric_part
        block_matrix = numpy.zeros((2 * self.n_qubits, 2 * self.n_qubits),
                                   dtype=complex)
        block_matrix[:self.n_qubits, :self.n_qubits] = antisymmetric_part
        block_matrix[:self.n_qubits, self.n_qubits:] = hermitian_part
        block_matrix[self.n_qubits:, :self.n_qubits] = -hermitian_part.conj()
        block_matrix[self.n_qubits:, self.n_qubits:] = (
            -antisymmetric_part.conj())

        ferm_unitary = self.quad_ham_npc.diagonalizing_bogoliubov_transform()

        # Check that the transformation is diagonalizing
        majorana_matrix, majorana_constant = self.quad_ham_npc.majorana_form()
        canonical, orthogonal = antisymmetric_canonical_form(majorana_matrix)
        diagonalized = ferm_unitary.conj().dot(
            block_matrix.dot(ferm_unitary.T.conj()))
        for i in numpy.ndindex((2 * self.n_qubits, 2 * self.n_qubits)):
            self.assertAlmostEqual(diagonalized[i], canonical[i])

        lower_unitary = ferm_unitary[self.n_qubits:]
        lower_left = lower_unitary[:, :self.n_qubits]
        lower_right = lower_unitary[:, self.n_qubits:]

        # Check that lower_left and lower_right satisfy the constraints
        # necessary for the transformed fermionic operators to satisfy
        # the fermionic anticommutation relations
        constraint_matrix_1 = (lower_left.dot(lower_left.T.conj()) +
                               lower_right.dot(lower_right.T.conj()))
        constraint_matrix_2 = (lower_left.dot(lower_right.T) +
                               lower_right.dot(lower_left.T))

        identity = numpy.eye(self.n_qubits, dtype=complex)
        for i in numpy.ndindex((self.n_qubits, self.n_qubits)):
            self.assertAlmostEqual(identity[i], constraint_matrix_1[i])
            self.assertAlmostEqual(0., constraint_matrix_2[i])


class AntisymmetricCanonicalFormTest(unittest.TestCase):

    def test_equality(self):
        """Test that the decomposition is valid."""
        n = 7
        rand_mat = numpy.random.randn(2 * n, 2 * n)
        antisymmetric_matrix = rand_mat - rand_mat.T
        canonical, orthogonal = antisymmetric_canonical_form(
            antisymmetric_matrix)
        result_matrix = orthogonal.dot(antisymmetric_matrix.dot(orthogonal.T))
        for i in numpy.ndindex(result_matrix.shape):
            self.assertAlmostEqual(result_matrix[i], canonical[i])

    def test_canonical(self):
        """Test that the returned canonical matrix has the right form."""
        n = 7
        # Obtain a random antisymmetric matrix
        rand_mat = numpy.random.randn(2 * n, 2 * n)
        antisymmetric_matrix = rand_mat - rand_mat.T
        canonical, orthogonal = antisymmetric_canonical_form(
            antisymmetric_matrix)
        for i in range(2 * n):
            for j in range(2 * n):
                if i < n and j == n + i:
                    self.assertTrue(canonical[i, j] > -EQ_TOLERANCE)
                elif i >= n and j == i - n:
                    self.assertTrue(canonical[i, j] < EQ_TOLERANCE)
                else:
                    self.assertAlmostEqual(canonical[i, j], 0.)

        diagonal = canonical[range(n), range(n, 2 * n)]
        for i in range(n - 1):
            self.assertTrue(diagonal[i] <= diagonal[i + 1])

    def test_bad_dimensions(self):
        n, p = (3, 4)
        ones_mat = numpy.ones((n, p))
        with self.assertRaises(ValueError):
            _ = antisymmetric_canonical_form(ones_mat)

    def test_not_antisymmetric(self):
        n = 4
        ones_mat = numpy.ones((n, n))
        with self.assertRaises(ValueError):
            _ = antisymmetric_canonical_form(ones_mat)



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

""" The transform function that does Fermion-qubit mappings
    based on a BinaryCode (arXiv:1712.07067) """

import numpy

from openfermion.ops import (BinaryCode,
                             FermionOperator,
                             QubitOperator,
                             BinaryPolynomial)


def extractor(binary_op):
    """ Applies the extraction superoperator to a binary expression
     to obtain the corresponding qubit operators.

    Args:
        binary_op (BinaryPolynomial): the binary term

    Returns (QubitOperator): the qubit operator corresponding to the
        binary terms
    """
    return_fn = 1
    for term in binary_op.terms:
        multiplier = 1
        if len(term) == 1:
            term = term[0]
            if isinstance(term, (numpy.int32, numpy.int64, int)):
                multiplier = QubitOperator('Z' + str(term))
            else:
                multiplier = -1
        elif len(term) > 1:
            multiplier = dissolve(term)

        return_fn *= multiplier
    return return_fn


def dissolve(term):
    """Decomposition helper. Takes a product of binary variables
    and outputs the Pauli-string sum that corresponds to the
    decomposed multi-qubit operator.

    Args:
        term (tuple): product of binary variables, i.e.: 'w0 w2 w3'

    Returns (QubitOperator): superposition of Pauli-strings

    Raises:
        ValueError: if the variable in term is not integer
    """
    prod = 2.0
    for var in term:
        if not isinstance(var, (numpy.int32, numpy.int64, int)):
            raise ValueError('dissolve only works on integers')
        prod *= (QubitOperator((), 0.5) - QubitOperator(
            'Z' + str(var), 0.5))
    return QubitOperator((), 1.0) - prod


def make_parity_list(code):
    """Create the parity list from the decoder of the input code.
    The output parity list has a similar structure as code.decoder. 

    Args:
        code (BinaryCode): the code to extract the parity list from.

    Returns (list): list of BinaryPolynomial, the parity list

    Raises:
        TypeError: argument is not BinaryCode
    """
    if not isinstance(code, BinaryCode):
        raise TypeError('argument is not BinaryCode')
    parity_binaries = [BinaryPolynomial()]
    for index in numpy.arange(code.n_modes - 1):
        parity_binaries += [parity_binaries[-1] + code.decoder[index]]
    return parity_binaries


def binary_code_transform(hamiltonian, code):
    """ Transforms a Hamiltonian written in fermionic basis into a Hamiltonian
    written in qubit basis, via a binary code.

    The role of the binary code is to relate the occupation vectors (v0 v1 v2
    ... vN-1) that span the fermionic basis, to the qubit basis, spanned by
    binary vectors (w0, w1, w2, ..., wn-1).

    The binary code has to provide an analytic relation between the binary
    vectors (v0, v1, ..., vN-1) and (w0, w1, ..., wn-1), and possibly has the
    property N>n, when the Fermion basis is smaller than the fermionic Fock
    space. The binary_code_transform function can transform Fermion operators
    to qubit operators for custom- and qubit-saving mappings.
    
    Note:
        Logic multi-qubit operators are decomposed into Pauli-strings (e.g.
        CPhase(1,2) = 0.5 * (1 + Z1 + Z2 - Z1 Z2 ) ), which might increase
        the number of Hamiltonian terms drastically.

    Args:
        hamiltonian (FermionOperator): the fermionic Hamiltonian
        code (BinaryCode): the binary code to transform the Hamiltonian

    Returns (QubitOperator): the transformed Hamiltonian

    Raises:
        TypeError: if the hamiltonian is not a FermionOperator or code is not
        a BinaryCode
    """
    if not isinstance(hamiltonian, FermionOperator):
        raise TypeError('hamiltonian provided must be a FermionOperator'
                        'received {}'.format(type(hamiltonian)))

    if not isinstance(code, BinaryCode):
        raise TypeError('code provided must be a BinaryCode'
                        'received {}'.format(type(code)))

    new_hamiltonian = QubitOperator()
    parity_list = make_parity_list(code)

    # for each term in hamiltonian
    for term, term_coefficient in hamiltonian.terms.items():

        """ the updated parity and occupation account for sign changes due
        changed occupations mid-way in the term """
        updated_parity = 0  # parity sign exponent
        parity_term = BinaryPolynomial()
        changed_occupation_vector = [0] * code.n_modes
        transformed_term = QubitOperator(())

        # keep track of indices appeared before
        fermionic_indices = numpy.array([])

        # for each multiplier
        for op_idx, op_tuple in enumerate(reversed(term)):
            # get count exponent, parity exponent addition
            fermionic_indices = numpy.append(fermionic_indices, op_tuple[0])
            count = numpy.count_nonzero(
                fermionic_indices[:op_idx] == op_tuple[0])
            updated_parity += numpy.count_nonzero(
                fermionic_indices[:op_idx] < op_tuple[0])

            # update term
            extracted = extractor(code.decoder[op_tuple[0]])
            extracted *= (((-1) ** count) * ((-1) ** (op_tuple[1])) * 0.5)
            transformed_term *= QubitOperator((), 0.5) - extracted

            # update parity term and occupation vector
            changed_occupation_vector[op_tuple[0]] += 1
            parity_term += parity_list[op_tuple[0]]

        # the parity sign and parity term
        transformed_term *= QubitOperator((), (-1) ** updated_parity)
        transformed_term *= extractor(parity_term)

        # the update operator
        changed_qubit_vector = numpy.mod(code.encoder.dot(
            changed_occupation_vector), 2)

        update_operator = QubitOperator(())
        for index, q_vec in enumerate(changed_qubit_vector):
            if q_vec:
                update_operator *= QubitOperator('X' + str(index))

        # append new term to new hamiltonian
        new_hamiltonian += term_coefficient * update_operator * \
                           transformed_term

    new_hamiltonian.compress()
    return new_hamiltonian


#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""symbolic binary class for decoder definitions (arxiv 1712.07067) """

import copy

import numpy

_SYMBOLIC_ONE = 'one'
_ACTION = 'W'


class BinaryPolynomialError(Exception):
    pass


class BinaryPolynomial(object):
    """The BinaryPolynomial class provides an analytic representation
    of non-linear binary functions. An instance of this class describes
    a term of binary variables (variables of the values {0,1}, indexed
    by integers like w0, w1, w2 and so on) that is considered to be evaluated
    modulo 2. This implies the following set of rules:
    
    the binary addition  w1 + w1 = 0, 
    binary multiplication w2 * w2 = w2 
    and power rule w3 ^ 0  = 1, where raising to every other
    integer power than zero reproduces w3.
    
    Of course, we can also add a non-trivial constant, which is 1.
    Due to these binary rules, every function available will be a 
    multinomial like e.g.
    
    1 + w1 w2 + w0 w1 .
    
    These binary functions are used for non-linear binary codes in order
    to decompress qubit bases back into fermion bases.  
    In that instance, one BinaryPolynomial object characterizes the occupation 
    of single orbital given a multi-qubit state in configuration 
    \|w0> \|w1> \|w2> ... . 
    
    For initialization, the preferred data types is either a string of the 
    multinomial, where each variable and constant is to be well separated by
    a whitespace, or in its native form of tuples,
    1 + w1 w2 + w0 w1 is represented as [(_SYMBOLIC_ONE,),(1,2),(0,1)]

    After initialization,BinaryPolynomial terms can be manipulated with the
     overloaded signs +, * and ^, according to the binary rules mentioned.

    Example:
        .. code-block:: python

            bin_fun = BinaryPolynomial('1 + w1 w2 + w0 w1')
            # Equivalently
            bin_fun = BinaryPolynomial(1) + BinaryPolynomial([(1,2),(0,1)])
            # Equivalently
            bin_fun = BinaryPolynomial([(_SYMBOLIC_ONE,),(1,2),(0,1)])

    Attributes:
        terms (list): a list of tuples. Each tuple represents a summand of the
            BinaryPolynomial term and each summand can contain multiple tuples
            representing the factors.
    """

    def __init__(self, term=None):
        """ Initialize the BinaryPolynomial based on term

        Args:
            term (str, list, tuple): used for initializing a BinaryPolynomial

        Raises:
            ValueError: when term is not a string,list or tuple
        """
        self.terms = []

        # long string input
        if isinstance(term, str) and '+' in term:
            self._long_string_init(term)
            return

        # Zero operator: leave the terms list empty
        if term is None:
            return

        # Sequence input: list of tuples of tuples
        elif isinstance(term, tuple) or isinstance(term, list):
            self._parse_sequence(list(term))

        # String input
        elif isinstance(term, str):
            self.terms.append(tuple(self._parse_string(term)))

        elif isinstance(term, (numpy.int32, numpy.int64, int)):
            # its a constant summand
            if term % 2:
                self._add_one()

        # Invalid input type
        else:
            raise ValueError('term specified incorrectly.')

        self._check_terms()

    def _check_terms(self):
        """ Ensures all terms obey binary rules, updates terms in place."""
        sorted_input = []
        for item in self.terms:
            if len(item):
                binary_sum_rule(sorted_input, tuple(set(item)))

        self.terms = sorted_input

    def _long_string_init(self, term):
        """Initialization from a long string representation i.e.:
        1 + w1 w2 + w3 w4'. Updates terms in place.

        Args:
            term (str): a string representation of a BinaryPolynomial
             that involves summation
        """
        for summand in term.split(' + '):
            # for each sum term
            parsed_summand = self._parse_string(summand)
            self.terms.append(parsed_summand)

        self._check_terms()

    @staticmethod
    def _check_factor(term, factor):
        """Checks types and values of all factors in a term,
        removes multiplications with 1.

        Args:
            term (list): a single term for BinaryPolynomial
            factor (int or str): a factor in term

        Returns (list): updated term

        Raises:
            ValueError: invalid action/negative or non-integer qubit index
        """

        term = list(term)

        if factor == _SYMBOLIC_ONE:
            if len(term) > 1:
                term.remove(factor)  # no need to keep 1 multiplier in the term
            return tuple(set(term))

        elif isinstance(factor, (numpy.int32, numpy.int64, int)):
            if factor < 0:
                raise ValueError('Invalid factor {},'
                                 'must be a positive integer'.format(factor))
            return tuple(set(term))
        else:
            raise ValueError('Invalid factor {}.'
                             'valid factor is positive integers and {} '
                             'for constant 1'.format(factor, _SYMBOLIC_ONE))

    def _parse_sequence(self, term):
        """ Parse a term given as a sequence type (i.e., list, tuple, etc.).
        e.g. [(0,1,2,3,_SYMBOLIC_ONE),...] -> [(0,1,2,3),...]. Updates terms
        in place

        Args:
            term (list): of tuples.
        """
        if term:
            for summand in term:
                for factor in summand:
                    summand = self._check_factor(summand, factor)
                if summand in self.terms:
                    self.terms.remove(summand)
                else:
                    self.terms.append(summand)
        else:
            self.terms = []

    @staticmethod
    def _parse_string(term):
        """ Parse a string term like 'w1 w2 w0'

        Args:
            term (str): string representation of BinaryPolynomial term.

        Returns (tuple): parsed string term

        Raises:
          ValueError: Incorrect terms
        """
        term_list = []
        add_one = False
        for factor in term.split():

            """ if 1 is already present; remove it since its not necessary to
            keep it if there are more than 1 terms"""
            if add_one:
                term_list.remove(_SYMBOLIC_ONE)  # if 1
                add_one = False

            if factor.isdigit():  # its a constant
                factor = int(factor) % 2
                if factor == 1:
                    # if there are other terms, no need to add another 1
                    if len(term_list) > 0:
                        continue
                    term_list.append(_SYMBOLIC_ONE)
                    add_one = True
                # multiply by zero
                elif factor == 0:
                    return []
            elif factor[1:].isdigit():
                q_idx = int(factor[1:])
                term_list.append(q_idx)
            else:
                raise ValueError('Invalid factor {}.'.format(factor))

        parsed_term = tuple(set(term_list))
        return parsed_term

    def enumerate_qubits(self):
        """ Enumerates all qubits indexed in a given BinaryPolynomial.

        Returns (list): a list of qubits
        """
        qubits = [factor for summand in self.terms
                  for factor in summand
                  if factor != _SYMBOLIC_ONE]

        return list(set(qubits))

    def shift(self, const):
        """ Shift all qubit indices by a given constant.

        Args:
            const (int): the constant to shift the indices by

        Raises:
            TypeError: const must be integer
        """
        if not isinstance(const, (numpy.int64, numpy.int32, int)):
            raise TypeError('can only shift qubit indices by an integer'
                            'received {}'.format(const))
        shifted_terms = []
        for summand in self.terms:
            shifted_summand = []
            for factor in summand:
                if factor != _SYMBOLIC_ONE:
                    shifted_summand.append(factor + const)
                else:
                    shifted_summand.append(factor)
            shifted_terms.append(tuple(set(shifted_summand)))

        self.terms = shifted_terms

    def evaluate(self, binary_list):
        """Evaluates a BinaryPolynomial

        Args:
            binary_list (list, array, str): a list of binary values
                corresponding  each binary variable 
                (in order of their indices) in the expression 

        Returns (int, 0 or 1): result of the evaluation

        Raises:
          BinaryPolynomialError: Length of list provided must match the number
                of qubits indexed in BinaryPolynomial
        """
        if isinstance(binary_list,str):
            binary_list = list(map(int, list(binary_list)))

        all_qubits = self.enumerate_qubits()
        if all_qubits:
            if max(all_qubits) >= len(binary_list):
                raise BinaryPolynomialError(
                    'the length of the binary list provided does not match'
                    ' the number of variables in the BinaryPolynomial')

            evaluation = 0
            for summand in self.terms:
                ev_tmp = 1
                for factor in summand:
                    if factor != _SYMBOLIC_ONE:
                        ev_tmp *= binary_list[factor]
                evaluation += ev_tmp
            return evaluation % 2

        elif self.terms:
            return 1
        else:
            return 0

    def _add_one(self):
        """ Adds constant 1 to a BinaryPolynomial. """

        # (_SYMBOLIC_ONE,) can only exist as a loner in BinaryPolynomial
        if (_SYMBOLIC_ONE,) in self.terms:
            self.terms.remove((_SYMBOLIC_ONE,))
        else:
            self.terms.append((_SYMBOLIC_ONE,))

    @classmethod
    def zero(cls):
        """
        Returns:
            additive_identity (BinaryPolynomial):
                A symbolic operator o with the property that o+x = x+o = x for
                all operators x of the same class.
        """
        return cls(term=[])

    @classmethod
    def identity(cls):
        """
        Returns:
            multiplicative_identity (BinaryPolynomial):
                A symbolic operator u with the property that u*x = x*u = x for
                all operators x of the same class.
        """
        return cls(term=[(_SYMBOLIC_ONE,)])

    def __str__(self):
        """ Return an easy-to-read string representation."""
        if not self.terms:
            return '0'
        string_rep = ''
        for term in self.terms:
            tmp_string = '['
            for factor in term:
                if factor == _SYMBOLIC_ONE:
                    tmp_string += '1 '
                else:
                    tmp_string += '{}{} '.format(_ACTION, factor)
            string_rep += '{}] + '.format(tmp_string.strip())
        return string_rep[:-3]

    def __repr__(self):
        return str(self)

    def __imul__(self, multiplier):
        """ In-place multiply (*=) with a scalar or operator of the same type.

        Args:
            multiplier(int or BinaryPolynomial): multiplier

        Returns:
            product (BinaryPolynomial): Mutated self.

        Raises:
          TypeError: Object of invalid type cannot multiply BinaryPolynomial.
        """

        # Handle integers.
        if isinstance(multiplier, (numpy.int64, numpy.int32, int)):
            mod_mul = int(multiplier % 2)
            if mod_mul:
                return self
            else:
                return self.zero()

        # Handle operator of the same type
        elif isinstance(multiplier, self.__class__):
            result_terms = []
            for left_term in self.terms:
                left_indices = set(
                    [term for term in left_term if term != _SYMBOLIC_ONE])

                for right_term in multiplier.terms:
                    right_indices = set(
                        [term for term in right_term if
                         term != _SYMBOLIC_ONE])

                    if len(left_indices) == 0 and len(right_indices) == 0:
                        product_term = (_SYMBOLIC_ONE,)
                        binary_sum_rule(result_terms, product_term)
                        continue

                    # binary rule - 2: w^2 = w
                    indices = left_indices | right_indices
                    product_term = sorted(list(indices))
                    binary_sum_rule(result_terms, tuple(product_term))

            self.terms = result_terms
            return self

        # Invalid multiplier type
        else:
            raise TypeError('Cannot multiply {} with {}'.format(
                self.__class__.__name__, multiplier.__class__.__name__))

    def __rmul__(self, multiplier):
        """ Return multiplier * self for a scalar or BinaryPolynomial.

        Args:
          multiplier (int or BinaryPolynomial): the multiplier of the
           BinaryPolynomial object

        Returns:
          product (BinaryPolynomial): A new instance of BinaryPolynomial.

        Raises:
          TypeError: Object of invalid type cannot multiply BinaryPolynomial.
        """
        if not isinstance(multiplier, (numpy.int64, numpy.int32, int,
                                       type(self))):
            raise TypeError(
                'Object of invalid type cannot multiply with ' +
                str(type(self)) + '.')
        return self * multiplier

    def __mul__(self, multiplier):
        """Return self * multiplier for int, or a BinaryPolynomial.

        Args:
            multiplier (int or BinaryPolynomial): the multiplier of the
           BinaryPolynomial object

        Returns:
            product (BinaryPolynomial): result of the multiplication

        Raises:
            TypeError: Invalid type cannot be multiply with BinaryPolynomial.
        """
        if isinstance(multiplier, (numpy.int64, numpy.int32, int,
                                   type(self))):
            product = copy.deepcopy(self)
            product *= multiplier
            return product
        else:
            raise TypeError(
                'Object of invalid type cannot multiply with ' +
                str(type(self)) + '.')

    def __iadd__(self, addend):
        """In-place method for += addition of a int or a BinaryPolynomial.

        Args:
            addend (int or BinaryPolynomial): The operator to add.

        Returns:
            sum (BinaryPolynomial): Mutated self.

        Raises:
            TypeError: Cannot add invalid type.
        """
        if isinstance(addend, type(self)):
            for term in addend.terms:
                binary_sum_rule(self.terms, term)
        if isinstance(addend, int):
            mod_add = addend % 2
            if mod_add:
                self._add_one()
        if not isinstance(addend, (numpy.int64, numpy.int32, int,
                                   type(self))):
            raise TypeError(
                'Object of invalid type cannot add with ' +
                str(type(self)) + '.')
        return self

    def __radd__(self, addend):
        """Method for right addition to BinaryPolynomial.

        Args:
            addend (int or BinaryPolynomial): The operator to add.

        Returns:
            sum (BinaryPolynomial): the sum of terms

        Raises:
            TypeError: Cannot add invalid type.
        """
        if not isinstance(addend, (numpy.int64, numpy.int32, int,
                                   type(self))):
            raise TypeError(
                'Object of invalid type cannot add with ' +
                str(type(self)) + '.')
        return self + addend

    def __add__(self, addend):
        """ 
        Left addition of BinaryPolynomial.
        Args:
            addend (int or BinaryPolynomial): The operator or int to add.

        Returns:
            sum (BinaryPolynomial): the sum of terms
        """
        summand = copy.deepcopy(self)
        summand += addend
        return summand

    def __pow__(self, exponent):
        """Exponentiate the BinaryPolynomial.

        Args:
            exponent (int): The exponent with which to raise the operator.

        Returns:
            exponentiated (BinaryPolynomial): Exponentiated symbolicBinary

        Raises:
            TypeError: Can only raise BinaryPolynomial to non-negative
                integer powers.
        """
        # Handle invalid exponents.
        if not isinstance(exponent, (numpy.int64, numpy.int32, int)):
            raise TypeError(
                'exponent must be int, but was {} {}'.format(
                    type(exponent), repr(exponent)))
        else:
            if exponent < 0:
                raise TypeError(
                    'exponent must be non-negative, but was {}'.format(
                        exponent))

        # Check if exponent is zero - if yes return self, if not return zero.
        if exponent == 0:
            return self.identity()
        else:
            return self


def binary_sum_rule(terms, summand):
    """ Updates terms in place based on binary rules.
    Args:
        terms: symbolicBinary terms
        summand: new potential addition to term
    """
    if summand not in terms:
        terms.append(summand)
    else:
        terms.remove(summand)



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Tests many modules to compute energy of LiH."""
from __future__ import absolute_import

import os
import numpy
import scipy.sparse
import unittest

from openfermion.config import *
from openfermion.hamiltonians import *
from openfermion.ops import *
from openfermion.transforms import *
from openfermion.utils import *


class LiHIntegrationTest(unittest.TestCase):

    def setUp(self):
        # Set up molecule.
        geometry = [('Li', (0., 0., 0.)), ('H', (0., 0., 1.45))]
        basis = 'sto-3g'
        multiplicity = 1
        filename = os.path.join(THIS_DIRECTORY, 'data',
                                'H1-Li1_sto-3g_singlet_1.45')
        self.molecule = MolecularData(
            geometry, basis, multiplicity, filename=filename)
        self.molecule.load()

        # Get molecular Hamiltonian.
        self.molecular_hamiltonian = self.molecule.get_molecular_hamiltonian()
        self.molecular_hamiltonian_no_core = (
            self.molecule.
            get_molecular_hamiltonian(occupied_indices=[0],
                                      active_indices=range(1,
                                                           self.molecule.
                                                           n_orbitals)))

        # Get FCI RDM.
        self.fci_rdm = self.molecule.get_molecular_rdm(use_fci=1)

        # Get explicit coefficients.
        self.nuclear_repulsion = self.molecular_hamiltonian.constant
        self.one_body = self.molecular_hamiltonian.one_body_tensor
        self.two_body = self.molecular_hamiltonian.two_body_tensor

        # Get fermion Hamiltonian.
        self.fermion_hamiltonian = normal_ordered(get_fermion_operator(
            self.molecular_hamiltonian))

        # Get qubit Hamiltonian.
        self.qubit_hamiltonian = jordan_wigner(self.fermion_hamiltonian)

        # Get explicit coefficients.
        self.nuclear_repulsion = self.molecular_hamiltonian.constant
        self.one_body = self.molecular_hamiltonian.one_body_tensor
        self.two_body = self.molecular_hamiltonian.two_body_tensor

        # Get matrix form.
        self.hamiltonian_matrix = get_sparse_operator(
            self.molecular_hamiltonian)
        self.hamiltonian_matrix_no_core = get_sparse_operator(
            self.molecular_hamiltonian_no_core)

    def test_all(self):

        # Test reverse Jordan-Wigner.
        fermion_hamiltonian = reverse_jordan_wigner(self.qubit_hamiltonian)
        fermion_hamiltonian = normal_ordered(fermion_hamiltonian)
        self.assertTrue(self.fermion_hamiltonian == fermion_hamiltonian)

        # Test mapping to interaction operator.
        fermion_hamiltonian = get_fermion_operator(self.molecular_hamiltonian)
        fermion_hamiltonian = normal_ordered(fermion_hamiltonian)
        self.assertTrue(self.fermion_hamiltonian == fermion_hamiltonian)

        # Test RDM energy.
        fci_rdm_energy = self.nuclear_repulsion
        fci_rdm_energy += numpy.sum(self.fci_rdm.one_body_tensor *
                                    self.one_body)
        fci_rdm_energy += numpy.sum(self.fci_rdm.two_body_tensor *
                                    self.two_body)
        self.assertAlmostEqual(fci_rdm_energy, self.molecule.fci_energy)

        # Confirm expectation on qubit Hamiltonian using reverse JW matches.
        qubit_rdm = self.fci_rdm.get_qubit_expectations(self.qubit_hamiltonian)
        qubit_energy = 0.0
        for term, coefficient in qubit_rdm.terms.items():
            qubit_energy += coefficient * self.qubit_hamiltonian.terms[term]
        self.assertAlmostEqual(qubit_energy, self.molecule.fci_energy)

        # Confirm fermionic RDMs can be built from measured qubit RDMs.
        new_fermi_rdm = get_interaction_rdm(qubit_rdm)
        fermi_rdm_energy = new_fermi_rdm.expectation(
            self.molecular_hamiltonian)
        self.assertAlmostEqual(fci_rdm_energy, self.molecule.fci_energy)

        # Test sparse matrices.
        energy, wavefunction = get_ground_state(self.hamiltonian_matrix)
        self.assertAlmostEqual(energy, self.molecule.fci_energy)
        expected_energy = expectation(self.hamiltonian_matrix, wavefunction)
        self.assertAlmostEqual(expected_energy, energy)

        # Make sure you can reproduce Hartree-Fock energy.
        hf_state = jw_hartree_fock_state(
            self.molecule.n_electrons, count_qubits(self.qubit_hamiltonian))
        hf_density = get_density_matrix([hf_state], [1.])
        expected_hf_density_energy = expectation(self.hamiltonian_matrix,
                                                 hf_density)
        expected_hf_energy = expectation(self.hamiltonian_matrix, hf_state)
        self.assertAlmostEqual(expected_hf_energy, self.molecule.hf_energy)
        self.assertAlmostEqual(expected_hf_density_energy,
                               self.molecule.hf_energy)

        # Check that frozen core result matches frozen core FCI from psi4.
        # Recore frozen core result from external calculation.
        self.frozen_core_fci_energy = -7.8807607374168
        no_core_fci_energy = scipy.linalg.eigh(
            self.hamiltonian_matrix_no_core.todense())[0][0]
        self.assertAlmostEqual(no_core_fci_energy,
                               self.frozen_core_fci_energy)

        # Check that the freeze_orbitals function has the same effect as the
        # as the occupied_indices option of get_molecular_hamiltonian.
        frozen_hamiltonian = freeze_orbitals(
                get_fermion_operator(self.molecular_hamiltonian), [0, 1])
        self.assertTrue(frozen_hamiltonian ==
                get_fermion_operator(self.molecular_hamiltonian_no_core))



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Tests for _bravyi_kitaev.py."""
from __future__ import absolute_import

import numpy
import unittest

from openfermion.ops import (FermionOperator,
                             QubitOperator)
from openfermion.transforms import (bravyi_kitaev,
                                    get_sparse_operator,
                                    jordan_wigner)
from openfermion.utils import eigenspectrum, number_operator


class BravyiKitaevTransformTest(unittest.TestCase):

    def test_bravyi_kitaev_transform(self):
        # Check that the QubitOperators are two-term.
        lowering = bravyi_kitaev(FermionOperator(((3, 0),)))
        raising = bravyi_kitaev(FermionOperator(((3, 1),)))
        self.assertEqual(len(raising.terms), 2)
        self.assertEqual(len(lowering.terms), 2)

        #  Test the locality invariant for N=2^d qubits
        # (c_j majorana is always log2N+1 local on qubits)
        n_qubits = 16
        invariant = numpy.log2(n_qubits) + 1
        for index in range(n_qubits):
            operator = bravyi_kitaev(FermionOperator(((index, 0),)), n_qubits)
            qubit_terms = operator.terms.items()  # Get the majorana terms.

            for item in qubit_terms:
                coeff = item[1]

                #  Identify the c majorana terms by real
                #  coefficients and check their length.
                if not isinstance(coeff, complex):
                    self.assertEqual(len(item[0]), invariant)

        #  Hardcoded coefficient test on 16 qubits
        lowering = bravyi_kitaev(FermionOperator(((9, 0),)), n_qubits)
        raising = bravyi_kitaev(FermionOperator(((9, 1),)), n_qubits)

        correct_operators_c = ((7, 'Z'), (8, 'Z'), (9, 'X'),
                               (11, 'X'), (15, 'X'))
        correct_operators_d = ((7, 'Z'), (9, 'Y'), (11, 'X'), (15, 'X'))

        self.assertEqual(lowering.terms[correct_operators_c], 0.5)
        self.assertEqual(lowering.terms[correct_operators_d], 0.5j)
        self.assertEqual(raising.terms[correct_operators_d], -0.5j)
        self.assertEqual(raising.terms[correct_operators_c], 0.5)

    def test_bk_identity(self):
        self.assertTrue(bravyi_kitaev(FermionOperator(())) ==
                        QubitOperator(()))

    def test_bk_n_qubits_too_small(self):
        with self.assertRaises(ValueError):
            bravyi_kitaev(FermionOperator('2^ 3^ 5 0'), n_qubits=4)

    def test_bk_jw_number_operator(self):
        # Check if number operator has the same spectrum in both
        # BK and JW representations
        n = number_operator(1, 0)
        jw_n = jordan_wigner(n)
        bk_n = bravyi_kitaev(n)

        # Diagonalize and make sure the spectra are the same.
        jw_spectrum = eigenspectrum(jw_n)
        bk_spectrum = eigenspectrum(bk_n)

        self.assertAlmostEqual(0., numpy.amax(
            numpy.absolute(jw_spectrum - bk_spectrum)))

    def test_bk_jw_number_operators(self):
        # Check if a number operator has the same spectrum in both
        # JW and BK representations
        n_qubits = 2
        n1 = number_operator(n_qubits, 0)
        n2 = number_operator(n_qubits, 1)
        n = n1 + n2

        jw_n = jordan_wigner(n)
        bk_n = bravyi_kitaev(n)

        # Diagonalize and make sure the spectra are the same.
        jw_spectrum = eigenspectrum(jw_n)
        bk_spectrum = eigenspectrum(bk_n)

        self.assertAlmostEqual(0., numpy.amax(
            numpy.absolute(jw_spectrum - bk_spectrum)))

    def test_bk_jw_number_operator_scaled(self):
        # Check if number operator has the same spectrum in both
        # JW and BK representations
        n_qubits = 1
        n = number_operator(n_qubits, 0, coefficient=2)  # eigenspectrum (0,2)
        jw_n = jordan_wigner(n)
        bk_n = bravyi_kitaev(n)

        # Diagonalize and make sure the spectra are the same.
        jw_spectrum = eigenspectrum(jw_n)
        bk_spectrum = eigenspectrum(bk_n)

        self.assertAlmostEqual(0., numpy.amax(
                               numpy.absolute(jw_spectrum - bk_spectrum)))

    def test_bk_jw_hopping_operator(self):
        # Check if the spectrum fits for a single hoppping operator
        n_qubits = 5
        ho = FermionOperator(((1, 1), (4, 0))) + FermionOperator(
            ((4, 1), (1, 0)))
        jw_ho = jordan_wigner(ho)
        bk_ho = bravyi_kitaev(ho)

        # Diagonalize and make sure the spectra are the same.
        jw_spectrum = eigenspectrum(jw_ho)
        bk_spectrum = eigenspectrum(bk_ho)

        self.assertAlmostEqual(0., numpy.amax(
                               numpy.absolute(jw_spectrum - bk_spectrum)))

    def test_bk_jw_majoranas(self):
        # Check if the Majorana operators have the same spectrum
        # irrespectively of the transform.
        n_qubits = 7

        a = FermionOperator(((1, 0),))
        a_dag = FermionOperator(((1, 1),))

        c = a + a_dag
        d = 1j * (a_dag - a)

        c_spins = [jordan_wigner(c), bravyi_kitaev(c)]
        d_spins = [jordan_wigner(d), bravyi_kitaev(d)]

        c_sparse = [get_sparse_operator(c_spins[0]),
                    get_sparse_operator(c_spins[1])]
        d_sparse = [get_sparse_operator(d_spins[0]),
                    get_sparse_operator(d_spins[1])]

        c_spectrum = [eigenspectrum(c_spins[0]),
                      eigenspectrum(c_spins[1])]
        d_spectrum = [eigenspectrum(d_spins[0]),
                      eigenspectrum(d_spins[1])]

        self.assertAlmostEqual(0., numpy.amax(numpy.absolute(d_spectrum[0] -
                                                             d_spectrum[1])))

    def test_bk_jw_integration(self):
        # This is a legacy test, which was a minimal failing example when
        # optimization for hermitian operators was used.
        n_qubits = 4

        # Minimal failing example:
        fo = FermionOperator(((3, 1),))

        jw = jordan_wigner(fo)
        bk = bravyi_kitaev(fo)

        jw_spectrum = eigenspectrum(jw)
        bk_spectrum = eigenspectrum(bk)

        self.assertAlmostEqual(0., numpy.amax(numpy.absolute(jw_spectrum -
                                                             bk_spectrum)))

    def test_bk_jw_integration_original(self):
        # This is a legacy test, which was an example proposed by Ryan,
        # failing when optimization for hermitian operators was used.
        n_qubits = 5
        fermion_operator = FermionOperator(((3, 1), (2, 1), (1, 0), (0, 0)),
                                           -4.3)
        fermion_operator += FermionOperator(((3, 1), (1, 0)), 8.17)
        fermion_operator += 3.2 * FermionOperator()

        # Map to qubits and compare matrix versions.
        jw_qubit_operator = jordan_wigner(fermion_operator)
        bk_qubit_operator = bravyi_kitaev(fermion_operator)

        # Diagonalize and make sure the spectra are the same.
        jw_spectrum = eigenspectrum(jw_qubit_operator)
        bk_spectrum = eigenspectrum(bk_qubit_operator)
        self.assertAlmostEqual(0., numpy.amax(numpy.absolute(jw_spectrum -
                                              bk_spectrum)), places=5)
